from airalogy.assigner import (
    AssignerBase,
    AssignerResult,
    assigner,
)
from airalogy.models.check import CheckValue


class RvAssigner(AssignerBase):
    @assigner(
        assigned_fields=[
            "rv_03",
            "rv_04",
        ],
        dependent_fields=[
            "rv_01",
            "rv_02",
        ],
        mode="auto",
    )
    def assign_rv03_and_04(dependent_fields: dict) -> AssignerResult:
        rv_01 = dependent_fields["rv_01"]
        rv_02 = dependent_fields["rv_02"]

        rv_03 = rv_01 + rv_02

        if rv_01 > rv_02:
            rv_04 = rv_01 - rv_02
            return AssignerResult(
                assigned_fields={
                    "rv_03": rv_03,
                    "rv_04": rv_04,
                },
            )
        else:
            return AssignerResult(
                success=False,
                error_message="rv_01 must be greater than rv_02",
            )

    @assigner(
        assigned_fields=[
            "rv_06",
        ],
        dependent_fields=[
            "rv_03",
            "rv_05",
        ],
        mode="manual",
    )
    def assign_rv06(dependent_fields: dict) -> AssignerResult:
        rv_03 = dependent_fields["rv_03"]
        rv_05 = dependent_fields["rv_05"]

        rv_06 = rv_03 - rv_05

        return AssignerResult(
            success=True,
            assigned_fields={
                "rv_06": rv_06,
            },
            error_message=None,
        )

    @assigner(
        assigned_fields=[
            "rv_07",
        ],
        dependent_fields=[
            "rv_04",
            "rv_06",
        ],
        mode="auto_force",
    )
    def assign_rv07(dependent_fields: dict) -> AssignerResult:
        rv_07 = dependent_fields["rv_04"] + dependent_fields["rv_07"]

        return AssignerResult(
            success=True,
            assigned_fields={
                "rv_07": rv_07,
            },
            error_message=None,
        )


class RvAssigner2(AssignerBase):
    @assigner(
        assigned_fields=[
            "rv_03",
            "rc_04",
        ],
        dependent_fields=[
            "rv_01",
            "rv_02",
            "rv_09",
        ],
        mode="auto",
    )
    def assign_rv03_and_04(dependent_fields: dict) -> AssignerResult:
        rv_03 = (
            dependent_fields["rv_01"]
            + dependent_fields["rv_02"]
            + dependent_fields["rv_09"]
        )
        rc_04 = (
            dependent_fields["rv_01"]
            + dependent_fields["rv_02"]
            - dependent_fields["rv_09"]
        )

        return AssignerResult(
            success=True,
            assigned_fields={
                "rv_03": rv_03,
                "rc_04": CheckValue(
                    checked=True,
                    annotation=f"rc_04 = rv_01 + rv_02 - rv_09, value is: {rc_04}",
                ),
            },
            error_message=None,
        )


def test_get_dependent_fields_of_assigned_key():
    assert RvAssigner.get_dependent_fields_of_assigned_key("rv_01") == []
    assert RvAssigner.get_dependent_fields_of_assigned_key("rv_02") == []
    assert sorted(RvAssigner.get_dependent_fields_of_assigned_key("rv_03")) == [
        "rv_01",
        "rv_02",
    ]
    assert sorted(RvAssigner.get_dependent_fields_of_assigned_key("rv_04")) == [
        "rv_01",
        "rv_02",
    ]
    assert sorted(RvAssigner.get_dependent_fields_of_assigned_key("rv_07")) == [
        "rv_01",
        "rv_02",
        "rv_03",
        "rv_04",
        "rv_05",
        "rv_06",
    ]

    assert RvAssigner2.get_dependent_fields_of_assigned_key("rv_01") == []
    assert RvAssigner2.get_dependent_fields_of_assigned_key("rv_02") == []
    assert sorted(RvAssigner2.get_dependent_fields_of_assigned_key("rv_03")) == [
        "rv_01",
        "rv_02",
        "rv_09",
    ]
    assert sorted(RvAssigner2.get_dependent_fields_of_assigned_key("rv_03")) == sorted(
        RvAssigner2.get_dependent_fields_of_assigned_key("rc_04")
    )


def test_get_assign_func_of_assigned_key():
    assert RvAssigner.get_assign_func_of_assigned_key("rv_01") is None
    assert RvAssigner.get_assign_func_of_assigned_key("rv_02") is None
    assert (
        RvAssigner.get_assign_func_of_assigned_key("rv_03").__name__
        == RvAssigner.assign_rv03_and_04.__name__
    )
    assert (
        RvAssigner.get_assign_func_of_assigned_key("rv_04").__name__
        == RvAssigner.assign_rv03_and_04.__name__
    )
    assert (
        RvAssigner.get_assign_func_of_assigned_key("rv_06").__name__
        == RvAssigner.assign_rv06.__name__
    )


def test_all_assigned_fields():
    rfs = RvAssigner.all_assigned_fields()
    assert sorted(rfs.keys()) == ["rv_03", "rv_04", "rv_06", "rv_07"]

    assert sorted(rfs["rv_03"]["dependent_fields"]) == ["rv_01", "rv_02"]
    assert rfs["rv_03"]["mode"] == "auto"

    assert sorted(rfs["rv_04"]["dependent_fields"]) == ["rv_01", "rv_02"]
    assert rfs["rv_04"]["mode"] == "auto"

    assert sorted(rfs["rv_06"]["dependent_fields"]) == [
        "rv_01",
        "rv_02",
        "rv_03",
        "rv_05",
    ]
    assert rfs["rv_06"]["mode"] == "manual"

    assert sorted(rfs["rv_07"]["dependent_fields"]) == [
        "rv_01",
        "rv_02",
        "rv_03",
        "rv_04",
        "rv_05",
        "rv_06",
    ]
    assert rfs["rv_07"]["mode"] == "auto_force"


def test_rv_assigner():
    dependent_data = {
        "rv_01": 20,
        "rv_02": 10,
        "rv_05": 15,
        "rv_09": 3,
    }

    result = RvAssigner.assign("rv_01", dependent_data)
    assert not result.success
    assert result.assigned_fields is None

    result = RvAssigner.assign("rv_03", dependent_data)
    assert result.success
    assert result.assigned_fields == {"rv_03": 30, "rv_04": 10}

    result = RvAssigner.assign("rv_03", {"rv_01": 10, "rv_02": 20})
    assert not result.success
    assert result.assigned_fields is None

    result = RvAssigner.assign("rv_04", dependent_data)
    assert result.success
    assert result.assigned_fields == {"rv_03": 30, "rv_04": 10}

    result = RvAssigner.assign("rv_06", dependent_data)
    assert result.success
    assert result.assigned_fields == {"rv_06": 15}

    result = RvAssigner2.assign("rv_03", dependent_data)
    assert result.success
    assert result.assigned_fields == {
        "rv_03": 33,
        "rc_04": CheckValue(
            checked=True, annotation="rc_04 = rv_01 + rv_02 - rv_09, value is: 27"
        ),
    }



import os

from airalogy.airalogy import Airalogy

os.environ["AIRALOGY_ENDPOINT"] = "http://localhost:4000"
os.environ["AIRALOGY_API_KEY"] = (
    "m0GqvQ/PqMgsaqWcykpP0Nx6f8+DrpoO/NVvLWKDcay1lPTCrjkdvsBJACBRR/jFyQyeBknS8FUV7UyDxxiwLYMxbh7U6H6KnScnEusaCF/F+WCoTgV5jrZPV0dw2pky"
)
os.environ["AIRALOGY_PROTOCOL_ID"] = (
    "airalogy.id.lab.lab1.project.proj1.protocol.test_node_sample.v.0.0.2"
)
airalogy = Airalogy()

# upload file
file = airalogy.upload_file_bytes("test.txt", b"test")
print(file)

# download file
print(airalogy.download_file_bytes(file["id"]))

# get file url
print(airalogy.get_file_url(file["id"]))

# download records
# print(
#     airalogy.download_records_json(
#         ["airalogy.id.record.01972062-22db-73f9-86f8-905515d3c88b.v.1"]
#     )
# )



"""
Define the model for the Aimd Step.
"""

from pydantic import BaseModel


class StepValue(BaseModel):
    """
    Model for the value of a Aimd Step.
    """

    checked: bool | None
    annotation: str



import re

html_comment_pattern = re.compile(r"<!--(.*?)-->")
var_name_pattern = re.compile(r"^[a-zA-Z][a-zA-Z0-9_]*$")

# one-line var pattern (previously rnv)
var_pattern = re.compile(r"\{\{var\|(.*?)\}\}")

# one-line var_table pattern (keep as is)
var_table_pattern = re.compile(r"\{\{var_table\|([^|]+?),\s*subvars=\[(.+?)\]\s*\}\}")
var_table_subvar_pattern = re.compile(
    r'subvar="([^"]+)",\s*subvar_title="([^"]+)",\s*subvar_explanation="([^"]+)"'
)
# multi-line var_table pattern
var_table_start_pattern = re.compile(r"\{\{var_table\|")
var_table_end_pattern = re.compile(r".*\}\}")

# one-line step pattern (previously rns)
step_pattern = re.compile(r"\{\{step\|(.*?)\}\}")
step_level_pattern = re.compile(r"^([1-9][0-9]?)$")
step_check_pattern = re.compile(r"check=(True|False)")
checked_message_pattern = re.compile(r'.*?,\s*checked_message="([^"]+)"')

# one-line check pattern (previously rnc)
check_pattern = re.compile(r"\{\{check\|(.*?)\}\}")


class AimdParseError(Exception):
    def __init__(self, message: str, line_number: int):
        super().__init__(f"{message} at line {line_number}.")
        self.line_number = line_number


def validate_var_name(type: str, var_name: str, line_number: int):
    if re.match(var_name_pattern, var_name):
        return True
    else:
        raise AimdParseError(f"Invalid {type} name definition: {var_name}", line_number)


def parse_step(step_str: str, line_number: int) -> dict:
    # step definition: {{step|<step_name>, <step_level>, check=True, checked_message="<message>"}}
    step = {"line_number": line_number, "level": 1}
    checked_message_match = re.match(checked_message_pattern, step_str)
    if checked_message_match is not None:
        step["checked_message"] = checked_message_match.group(1)

    args = [str.strip() for str in step_str.split(",")]
    if len(args) < 1:
        raise AimdParseError("Invalid step definition", line_number)

    step_name = args[0]
    validate_var_name("step", step_name, line_number)
    step["name"] = step_name

    current_arg_index = 1
    if len(args) >= 2 and re.match(step_level_pattern, args[current_arg_index]):
        step["level"] = int(args[current_arg_index])
        current_arg_index += 1

    if current_arg_index < len(args):
        check_match = re.match(step_check_pattern, args[current_arg_index])
        if check_match is None:
            raise AimdParseError("Invalid step check definition", line_number)
        step["check"] = check_match.group(1) == "True"

    return step


def parse_check(check_str: str, line_number: int) -> dict:
    # check definition: {{check|<check_name>, checked_message="<message>"}}
    check = {"line_number": line_number}
    checked_message_match = re.match(checked_message_pattern, check_str)
    if checked_message_match is not None:
        check["checked_message"] = checked_message_match.group(1)

    args = [str.strip() for str in check_str.split(",")]
    if len(args) < 1:
        raise AimdParseError("Invalid check definition", line_number)

    check_name = args[0]
    validate_var_name("check", check_name, line_number)
    check["name"] = check_name
    return check


def parse_variable(var_name: str, line_number: int) -> dict:
    # variable definition: {{var|<variable_name>}}
    var = {"line_number": line_number}
    validate_var_name("var", var_name, line_number)
    var["name"] = var_name
    return var


def parse_var_table(rv_str_matches: tuple[str, str], line_number: int) -> dict:
    # research_variable_table definition: {{var_table|<var_table_name>, subvars=[<subvar_name_1>, <subvar_name_2>, ...]}}
    rv_name, subvars_str = rv_str_matches
    validate_var_name("var_table", rv_name, line_number)
    subvars = []
    for str in subvars_str.split(","):
        subvar_name = str.strip()
        validate_var_name("subvar", subvar_name, line_number)
        subvars.append(subvar_name)
    return {
        "name": rv_name,
        "subvars": subvars,
        "type": "table",
        "line_number": line_number,
    }


def validate_var_name_unique(
    steps: list[dict], variables: list[dict], checks: list[dict]
):
    step_names = set()
    [step_names.add(step["name"]) for step in steps]
    var_names = set()
    for var in variables:
        if var["name"] in step_names:
            raise AimdParseError(
                f"Duplicate variable name: {var['name']}", var["line_number"]
            )
        var_names.add(var["name"])
    for check in checks:
        if check["name"] in step_names or check["name"] in var_names:
            raise AimdParseError(
                f"Duplicate variable name: {check['name']}", check["line_number"]
            )
    return True


# extract variable from aimd file
def extract_vars(aimd_content: str) -> dict:
    steps = []
    vars = []
    checks = []
    line_number = 0
    in_block = False
    block_start_line_number = 0
    block_lines = ""
    lines = aimd_content.splitlines()
    for line in lines:
        line = line.strip()
        line_number += 1
        # skip comment
        if line == "" or re.match(html_comment_pattern, line):
            continue
        # multi-line var_table
        if (
            re.match(var_table_start_pattern, line) is not None
            and re.match(var_table_end_pattern, line) is None
        ):
            in_block = True
            block_start_line_number = line_number
            block_lines = line
            continue
        if in_block:
            if re.match(var_table_start_pattern, line) is not None:
                raise AimdParseError(
                    "Nested var_table block not supported", line_number
                )

            block_lines += line + " "
            if re.match(var_table_end_pattern, line) is not None:
                in_block = False
                match = re.match(var_table_pattern, block_lines)
                if match is not None:
                    vars.append(
                        parse_var_table(match.groups(), block_start_line_number)
                    )
                block_lines = ""
            continue
        # step
        step_matches = re.findall(step_pattern, line)
        step_matches_len = len(step_matches)
        if step_matches_len > 1:
            raise AimdParseError("Invalid step definition", line_number)
        if step_matches_len == 1:
            step = parse_step(step_matches[0], line_number)
            steps.append(step)
        # var_table
        var_table_matches = re.findall(var_table_pattern, line)
        if len(var_table_matches) > 0:
            [
                vars.append(parse_var_table(var, line_number))
                for var in var_table_matches
            ]
        # variable
        var_matches = re.findall(var_pattern, line)
        if len(var_matches) > 0:
            [vars.append(parse_variable(var, line_number)) for var in var_matches]
        # check
        check_matches = re.findall(check_pattern, line)
        if len(check_matches) > 0:
            [checks.append(parse_check(check, line_number)) for check in check_matches]
    validate_var_name_unique(steps, vars, checks)
    return {
        "steps": steps,
        "vars": vars,
        "checks": checks,
    }



import json
import hashlib

from airalogy.record.hash import get_data_sha1


def compute_expected_sha1(data: dict) -> str:
    """
    Compute the expected SHA-1 hash for the given data using the same serialization rules.
    """
    data_str = json.dumps(
        data,
        sort_keys=True,
        separators=(",", ":"),
        ensure_ascii=False,
    )
    return hashlib.sha1(data_str.encode("utf-8")).hexdigest()


def test_empty_data():
    """
    Test the case when 'data' is an empty dictionary.
    """
    record = {"data": {}}
    expected = compute_expected_sha1({})
    assert get_data_sha1(record) == expected


def test_simple_data():
    """
    Test the case when 'data' is a simple dictionary.
    """
    data = {"name": "Alice爱丽丝", "age": 30}
    record = {"data": data}
    expected = compute_expected_sha1(data)
    assert get_data_sha1(record, print_data_str=True) == expected


def test_nested_data():
    """
    Test the case when 'data' contains nested structures.
    """
    data = {
        "user": {"name": "Bob", "languages": ["Python", "JavaScript"]},
        "active": True,
    }
    record = {"data": data}
    expected = compute_expected_sha1(data)
    assert get_data_sha1(record) == expected



from airalogy.assigner import (
    AssignerBase,
    AssignerResult,
    assigner,
)


class RvAssigner(AssignerBase):
    @assigner(
        assigned_fields=[
            "rv_table_1.subrv_3",
        ],
        dependent_fields=[
            "rv_01",
            "rv_table_1.subrv_1",
            "rv_table_1.subrv_2",
        ],
        mode="auto",
    )
    def assign_rv_table(dependent_fields: dict) -> AssignerResult:
        subrv_3 = (
            dependent_fields["rv_table_1.subrv_1"]
            + dependent_fields["rv_table_1.subrv_2"]
            + dependent_fields["rv_01"]
        )

        return AssignerResult(
            success=True,
            assigned_fields={
                "rv_table_1.subrv_3": subrv_3,
            },
            error_message=None,
        )

    @assigner(
        assigned_fields=[
            "rv_02",
        ],
        dependent_fields=[
            "rv_01",
            "rv_table_1",
        ],
        mode="auto",
    )
    def assign_rv_02(dependent_fields: dict) -> AssignerResult:
        sum = 0
        for row in dependent_fields["rv_table_1"]:
            sum += row["subrv_1"] + row["subrv_2"]
        rv_02 = dependent_fields["rv_01"] + sum

        return AssignerResult(
            success=True,
            assigned_fields={
                "rv_02": rv_02,
            },
            error_message=None,
        )


def test_rv_table_assigner():
    rfs = RvAssigner.all_assigned_fields()

    assert sorted(rfs.keys()) == ["rv_02", "rv_table_1.subrv_3"]

    assert sorted(rfs["rv_table_1.subrv_3"]["dependent_fields"]) == [
        "rv_01",
        "rv_table_1.subrv_1",
        "rv_table_1.subrv_2",
    ]
    assert sorted(rfs["rv_02"]["dependent_fields"]) == ["rv_01", "rv_table_1"]

    result = RvAssigner.assign(
        "rv_table_1.subrv_3",
        {"rv_01": 10, "rv_table_1.subrv_1": 20, "rv_table_1.subrv_2": 30},
    )
    assert result.success
    assert result.assigned_fields == {"rv_table_1.subrv_3": 60}

    result = RvAssigner.assign(
        "rv_02",
        {
            "rv_01": 10,
            "rv_table_1": [
                {"subrv_1": 10, "subrv_2": 20, "subrv_3": 60},
                {"subrv_1": 15, "subrv_2": 10, "subrv_3": 35},
            ],
        },
    )
    assert result.success
    assert result.assigned_fields == {"rv_02": 65}



import pytest
from airalogy.types import ATCG


def test_valid_atcg():
    s = ATCG("ATCGGATC")
    assert str(s) == "ATCGGATC"


def test_invalid_atcg():
    with pytest.raises(ValueError):
        ATCG("ATCX")
    with pytest.raises(ValueError):
        ATCG("atcg")  # lower case not allowed
    with pytest.raises(TypeError):
        ATCG(123)


def test_complement():
    s = ATCG("ATCG")
    assert s.complement() == "TAGC"
    s2 = ATCG("AATTCCGG")
    assert s2.complement() == "TTAAGGCC"


def test_pydantic_model_with_atcg():
    from pydantic import BaseModel, ValidationError

    class DNA(BaseModel):
        seq: ATCG

    # Valid case
    m = DNA(seq="ATCGGATC")
    assert isinstance(m.seq, ATCG)
    assert m.seq == "ATCGGATC"

    # Invalid case
    with pytest.raises(ValidationError):
        DNA(seq="ATCX")
    with pytest.raises(ValidationError):
        DNA(seq=123)

    # Schema should contain airalogy_built_in_type
    schema = DNA.model_json_schema()
    assert schema["properties"]["seq"]["airalogy_type"] == "ATCG"



"""
Define the model for the Checkpoint Value.
"""

from pydantic import BaseModel


class CheckValue(BaseModel):
    """
    Model for the value of a Checkpoint.
    """

    checked: bool
    annotation: str



import pytest
from airalogy.types import VersionStr, SnakeStr, ProtocolId, RecordId


class TestVersionStr:
    """Tests for VersionStr type"""

    def test_valid_versions(self):
        assert VersionStr("1.2.3") == "1.2.3"
        assert isinstance(VersionStr("0.0.1"), VersionStr)
        assert VersionStr("10.20.30") == "10.20.30"
        assert isinstance(VersionStr("999.999.999"), VersionStr)

    def test_invalid_versions(self):
        with pytest.raises(ValueError):
            VersionStr("1.2")  # missing patch
        with pytest.raises(ValueError):
            VersionStr("a.b.c")  # non-numeric
        with pytest.raises(ValueError):
            VersionStr("1.2.3.4")  # too many parts
        with pytest.raises(ValueError):
            VersionStr("")  # empty string
        with pytest.raises(ValueError):
            VersionStr("1.2.3-alpha")  # with suffix

    def test_type_annotations(self):
        """Test that type annotations work correctly for VersionStr"""
        version: VersionStr = VersionStr("1.2.3")

        # Should work as normal strings
        assert version.replace(".", "-") == "1-2-3"
        assert len(version) == 5

        # Type checking should work (runtime)
        assert isinstance(version, VersionStr)
        assert isinstance(version, str)

    def test_pydantic_integration(self):
        """Test VersionStr with Pydantic models"""
        from pydantic import BaseModel, ValidationError

        class Model(BaseModel):
            version: VersionStr

        # Valid case
        m = Model(version="1.2.3")
        assert isinstance(m.version, VersionStr)
        assert m.version == "1.2.3"

        # Invalid case
        with pytest.raises(ValidationError):
            Model(version="invalid.version")

        # Schema should contain airalogy_type
        schema = Model.model_json_schema()
        assert schema["properties"]["version"]["airalogy_type"] == "VersionStr"
        assert schema["properties"]["version"]["pattern"] == r"^\d+\.\d+\.\d+$"


class TestSnakeStr:
    """Tests for SnakeStr type"""

    def test_valid_snake_case(self):
        assert SnakeStr("snake_case") == "snake_case"
        assert SnakeStr("snake_case_123") == "snake_case_123"
        assert isinstance(SnakeStr("snake1_case2"), SnakeStr)
        # Edge cases - valid
        assert SnakeStr("a") == "a"  # single letter
        assert SnakeStr("a1") == "a1"  # letter + digit
        assert SnakeStr("a1b2c3") == "a1b2c3"  # mixed letters and digits
        assert (
            SnakeStr("test_123_abc") == "test_123_abc"
        )  # multiple segments with digits
        assert SnakeStr("a_b_c_d_e") == "a_b_c_d_e"  # many segments

    def test_invalid_snake_case(self):
        with pytest.raises(ValueError):
            SnakeStr("CamelCase")  # camel case
        with pytest.raises(ValueError):
            SnakeStr("snake-case")  # hyphen instead of underscore
        with pytest.raises(ValueError):
            SnakeStr("snake__case")  # consecutive underscores
        with pytest.raises(ValueError):
            SnakeStr("snakeCase")  # mixed case

    def test_edge_cases_invalid(self):
        """Test edge cases that should be invalid"""
        with pytest.raises(ValueError):
            SnakeStr("")  # empty string
        with pytest.raises(ValueError):
            SnakeStr("1snake")  # starts with digit
        with pytest.raises(ValueError):
            SnakeStr("_snake")  # starts with underscore
        with pytest.raises(ValueError):
            SnakeStr("a_")  # ends with underscore
        with pytest.raises(ValueError):
            SnakeStr("snake_")  # ends with underscore
        with pytest.raises(ValueError):
            SnakeStr("snake___case")  # multiple consecutive underscores
        with pytest.raises(ValueError):
            SnakeStr("SNAKE_CASE")  # uppercase letters
        with pytest.raises(ValueError):
            SnakeStr("snake case")  # space
        with pytest.raises(ValueError):
            SnakeStr("snake.case")  # dot
        with pytest.raises(ValueError):
            SnakeStr("snake@case")  # special character

    def test_type_annotations(self):
        """Test that type annotations work correctly for SnakeStr"""
        a_str: SnakeStr = SnakeStr("snake_case")

        # Should work as normal strings
        assert a_str.upper() == "SNAKE_CASE"
        assert len(a_str) == 10

        # Type checking should work (runtime)
        assert isinstance(a_str, SnakeStr)
        assert isinstance(a_str, str)

    def test_pydantic_integration(self):
        """Test SnakeStr with Pydantic models"""
        from pydantic import BaseModel, ValidationError

        class Model(BaseModel):
            name: SnakeStr

        # Valid case
        m = Model(name="snake_case")
        assert isinstance(m.name, SnakeStr)
        assert m.name == "snake_case"

        # Invalid case
        with pytest.raises(ValidationError):
            Model(name="CamelCase")

        # Schema should contain airalogy_type
        schema = Model.model_json_schema()
        assert schema["properties"]["name"]["airalogy_type"] == "SnakeStr"


class TestProtocolId:
    """Tests for ProtocolId type"""

    def test_valid_protocol_ids(self):
        valid_id = (
            "airalogy.id.lab.my_lab.project.test_project.protocol.data_analysis.v.1.2.3"
        )
        protocol_id = ProtocolId(valid_id)
        assert protocol_id == valid_id
        assert isinstance(protocol_id, ProtocolId)

        # Test with numbers in components
        valid_id2 = "airalogy.id.lab.lab1.project.proj2.protocol.proto3.v.0.0.1"
        protocol_id2 = ProtocolId(valid_id2)
        assert protocol_id2 == valid_id2

    def test_create_method(self):
        protocol_id = ProtocolId.create(
            "my_lab", "test_project", "data_analysis", "1.2.3"
        )
        expected = (
            "airalogy.id.lab.my_lab.project.test_project.protocol.data_analysis.v.1.2.3"
        )
        assert protocol_id == expected

    def test_invalid_protocol_ids(self):
        # Wrong format
        with pytest.raises(ValueError):
            ProtocolId("wrong.format")

        # Invalid snake_case components
        with pytest.raises(ValueError):
            ProtocolId.create("Invalid-Lab", "test_project", "data_analysis", "1.2.3")

        # Invalid version
        with pytest.raises(ValueError):
            ProtocolId.create(
                "my_lab", "test_project", "data_analysis", "invalid.version"
            )

        # Consecutive underscores
        with pytest.raises(ValueError):
            ProtocolId(
                "airalogy.id.lab.my__lab.project.test_project.protocol.data_analysis.v.1.2.3"
            )

    def test_pydantic_integration(self):
        """Test ProtocolId with Pydantic models"""
        from pydantic import BaseModel, ValidationError

        class Model(BaseModel):
            protocol_id: ProtocolId

        # Valid case
        valid_id = (
            "airalogy.id.lab.my_lab.project.test_project.protocol.data_analysis.v.1.2.3"
        )
        m = Model(protocol_id=valid_id)
        assert isinstance(m.protocol_id, ProtocolId)
        assert m.protocol_id == valid_id

        # Invalid case
        with pytest.raises(ValidationError):
            Model(protocol_id="invalid.protocol.id")

        # Schema should contain airalogy_type
        schema = Model.model_json_schema()
        assert schema["properties"]["protocol_id"]["airalogy_type"] == "ProtocolId"


class TestRecordId:
    """Tests for RecordId type"""

    def test_valid_record_ids(self):
        valid_id = "airalogy.id.record.550e8400-e29b-41d4-a716-446655440000.v.1"
        record_id = RecordId(valid_id)
        assert record_id == valid_id
        assert isinstance(record_id, RecordId)

        # Test with higher version
        valid_id2 = "airalogy.id.record.550e8400-e29b-41d4-a716-446655440000.v.999"
        record_id2 = RecordId(valid_id2)
        assert record_id2 == valid_id2

    def test_create_method(self):
        uuid_str = "550e8400-e29b-41d4-a716-446655440000"
        record_id = RecordId.create(uuid_str, 1)
        expected = f"airalogy.id.record.{uuid_str}.v.1"
        assert record_id == expected

        # Test with higher version
        record_id2 = RecordId.create(uuid_str, 999)
        expected2 = f"airalogy.id.record.{uuid_str}.v.999"
        assert record_id2 == expected2

    def test_create_method_invalid(self):
        """Test invalid inputs to create method"""
        # Invalid version in create method
        with pytest.raises(ValueError):
            RecordId.create("550e8400-e29b-41d4-a716-446655440000", 0)

        # Invalid UUID in create method
        with pytest.raises(ValueError):
            RecordId.create("invalid-uuid", 1)

        # Negative version
        with pytest.raises(ValueError):
            RecordId.create("550e8400-e29b-41d4-a716-446655440000", -1)

    def test_invalid_record_ids(self):
        # Wrong format
        with pytest.raises(ValueError):
            RecordId("wrong.format")

        # Invalid UUID
        with pytest.raises(ValueError):
            RecordId("airalogy.id.record.invalid-uuid.v.1")

        # Version < 1
        with pytest.raises(ValueError):
            RecordId("airalogy.id.record.550e8400-e29b-41d4-a716-446655440000.v.0")

        # Negative version
        with pytest.raises(ValueError):
            RecordId("airalogy.id.record.550e8400-e29b-41d4-a716-446655440000.v.-1")

    def test_pydantic_integration(self):
        """Test RecordId with Pydantic models"""
        from pydantic import BaseModel, ValidationError

        class Model(BaseModel):
            record_id: RecordId

        # Valid case
        valid_id = "airalogy.id.record.550e8400-e29b-41d4-a716-446655440000.v.1"
        m = Model(record_id=valid_id)
        assert isinstance(m.record_id, RecordId)
        assert m.record_id == valid_id

        # Invalid case
        with pytest.raises(ValidationError):
            Model(record_id="invalid.record.id")

        # Schema should contain airalogy_type
        schema = Model.model_json_schema()
        assert schema["properties"]["record_id"]["airalogy_type"] == "RecordId"
        assert (
            schema["properties"]["record_id"]["pattern"]
            == r"^airalogy\.id\.record\.([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})\.v\.(\d+)$"
        )



"""
Built-in type: CurrentTime
"""

from datetime import datetime

from pydantic import Field
from typing_extensions import Annotated

_ISO_PATTERN = r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d+)?(?:Z|[+-]\d{2}:\d{2})$"

CurrentTime = Annotated[
    datetime,
    Field(
        description="The current time. The timezone is based on the user's browser.",
        json_schema_extra={"airalogy_type": "CurrentTime"},
        pattern=_ISO_PATTERN,
    ),
]



import pytest

from airalogy.aimd.parser import AimdParseError, extract_vars


def test_extract_vars():
    excepted_result = {
        "steps": [
            {"line_number": 10, "name": "step_1", "level": 1},
            {"line_number": 11, "name": "step_1_1", "level": 2},
            {
                "line_number": 12,
                "name": "step_1_2",
                "level": 2,
                "check": True,
            },
            {
                "line_number": 13,
                "checked_message": "The next step is very important, please be careful.",
                "name": "step_1_3",
                "level": 2,
                "check": True,
            },
            {"line_number": 15, "name": "step_2", "level": 1},
            {"line_number": 39, "name": "step_3", "level": 1, "check": True},
            {"line_number": 40, "name": "step_3_1", "level": 2},
            {"line_number": 41, "name": "step_3_2", "level": 2},
        ],
        "vars": [
            {"line_number": 5, "name": "user"},
            {"line_number": 6, "name": "current_time"},
            {"line_number": 18, "name": "rv_image"},
            {"line_number": 21, "name": "rv_image_flip"},
            {"line_number": 24, "name": "rv_enum_01"},
            {"line_number": 30, "name": "pcr_temp_denaturation_pre"},
            {"line_number": 30, "name": "pcr_time_denaturation_pre"},
            {"line_number": 31, "name": "pcr_temp_denaturation"},
            {"line_number": 31, "name": "pcr_time_denaturation"},
            {"line_number": 32, "name": "pcr_temp_annealing"},
            {"line_number": 32, "name": "pcr_time_annealing"},
            {"line_number": 33, "name": "pcr_temp_extension"},
            {"line_number": 33, "name": "pcr_time_extension"},
            {"line_number": 34, "name": "pcr_temp_extension_post"},
            {"line_number": 34, "name": "pcr_time_extension_post"},
            {"line_number": 35, "name": "pcr_temp_keep"},
            {"line_number": 35, "name": "pcr_time_keep"},
            {"line_number": 37, "name": "pcr_cycle"},
            {
                "name": "pcr_dna_groups",
                "subvars": [
                    "pcr_identifier",
                    "forward_primer_sequence",
                    "reverse_primer_sequence",
                    "template_dna_sequence",
                    "template_dna_concentration",
                ],
                "type": "table",
                "line_number": 44,
            },
            {
                "name": "pcr_dna_groups_2",
                "subvars": [
                    "pcr_identifier",
                    "forward_primer_sequence",
                    "reverse_primer_sequence",
                    "template_dna_sequence",
                    "template_dna_concentration",
                ],
                "type": "table",
                "line_number": 47,
            },
            {"line_number": 52, "name": "var_01"},
            {"line_number": 53, "name": "var_02"},
            {"line_number": 54, "name": "var_03"},
            {"line_number": 55, "name": "var_04"},
            {"line_number": 56, "name": "var_05"},
            {"line_number": 57, "name": "var_06"},
            {"line_number": 66, "name": "pcr_result"},
        ],
        "checks": [
            {"line_number": 60, "name": "check_1"},
            {"line_number": 61, "name": "check_2"},
            {
                "line_number": 62,
                "checked_message": "Please be careful not to let the condensation on the PCR tube wall drop into the PCR reaction system.",
                "name": "check_3",
            },
        ],
    }
    with open("tests/test_aimd/test.aimd") as f:
        aimd_content = f.read()
        result = extract_vars(aimd_content)
        assert result == excepted_result


def test_extract_vars_invalid_step_definition():
    aimd_content = """
        {{step|Step 1, 1, check=True, checked_message="Step 1 is checked"}}
        {{step|Step 2, 2, check=True, checked_message="Step 2 is checked"}}
    """

    with pytest.raises(AimdParseError):
        extract_vars(aimd_content)


def test_extract_vars_duplicate_variable_name():
    aimd_content = """
        {{step|Step 1, 1, check=True, checked_message="Step 1 is checked"}}
        {{var|Variable 1, input_type=image}}
        {{var|Variable 1, input_type=audio}}
    """

    with pytest.raises(AimdParseError):
        extract_vars(aimd_content)




from airalogy.types import ATCG, CurrentTime, UserName
from pydantic import BaseModel

class DNA(BaseModel):
    seq: ATCG
    time: CurrentTime
    user: UserName

# m = DNA(seq="ATCGGATC")

s1 = DNA.model_json_schema()

print(s1)



from airalogy.record.validator import (
    all_var_ids_in_records,
)

import pytest


def test_all_var_ids_present():
    """
    Test when all records contain all the specified variable IDs.
    """
    records = [
        {
            "data": {
                "var": {
                    "var1": "value1",
                    "var2": "value2",
                }
            }
        },
        {
            "data": {
                "var": {
                    "var1": "value3",
                    "var2": "value4",
                }
            }
        },
    ]
    var_ids = ["var1", "var2"]
    assert all_var_ids_in_records(records, var_ids) is True


def test_some_var_ids_missing():
    """
    Test when some records are missing some of the specified variable IDs.
    """
    records = [
        {
            "data": {
                "var": {
                    "var1": "value1",
                }
            }
        },
        {
            "data": {
                "var": {
                    "var1": "value3",
                    "var2": "value4",
                }
            }
        },
    ]
    var_ids = ["var1", "var2"]
    assert all_var_ids_in_records(records, var_ids) is False


def test_empty_var_ids():
    """
    Test when the var_ids list is empty. Should raise a ValueError.
    """
    records = [
        {
            "data": {
                "var": {
                    "var1": "value1",
                }
            }
        },
        {
            "data": {
                "var": {
                    "var2": "value4",
                }
            }
        },
    ]
    var_ids = []
    with pytest.raises(ValueError):
        all_var_ids_in_records(records, var_ids)


def test_empty_records():
    """
    Test when the records list is empty. Should raise a ValueError.
    """
    records = []
    var_ids = ["var1", "var2"]
    with pytest.raises(ValueError):
        all_var_ids_in_records(records, var_ids)


def test_missing_data_key():
    """
    Test when a record is missing the 'data' key.
    """
    records = [
        {
            # 'data' key is missing
        },
        {
            "data": {
                "var": {
                    "var1": "value3",
                    "var2": "value4",
                }
            }
        },
    ]
    var_ids = ["var1", "var2"]
    with pytest.raises(ValueError):
        all_var_ids_in_records(records, var_ids)


def test_missing_var_key():
    """
    Test when a record's 'data' dictionary is missing the 'var' key.
    """
    records = [
        {
            "data": {
                # 'var' key is missing
            }
        },
        {
            "data": {
                "var": {
                    "var1": "value3",
                    "var2": "value4",
                }
            }
        },
    ]
    var_ids = ["var1", "var2"]
    with pytest.raises(ValueError):
        all_var_ids_in_records(records, var_ids)


def test_var_is_not_dict():
    """
    Test when the 'var' key in 'data' is not a dictionary.
    """
    records = [
        {"data": {"var": "not a dict"}},
        {
            "data": {
                "var": {
                    "var1": "value3",
                    "var2": "value4",
                }
            }
        },
    ]
    var_ids = ["var1", "var2"]
    with pytest.raises(ValueError):
        all_var_ids_in_records(records, var_ids)


def test_data_is_not_dict():
    """
    Test when the 'data' key is not a dictionary.
    """
    records = [
        {"data": "not a dict"},
        {
            "data": {
                "var": {
                    "var1": "value3",
                    "var2": "value4",
                }
            }
        },
    ]
    var_ids = ["var1", "var2"]
    with pytest.raises(ValueError):
        all_var_ids_in_records(records, var_ids)


def test_var_contains_extra_ids():
    """
    Test when 'var' contains extra variable IDs beyond those specified.
    Should still return True as long as specified IDs are present.
    """
    records = [
        {
            "data": {
                "var": {
                    "var1": "value1",
                    "var2": "value2",
                    "var_extra": "value_extra",
                }
            }
        },
        {
            "data": {
                "var": {
                    "var1": "value3",
                    "var2": "value4",
                    "var_extra": "value_extra",
                }
            }
        },
    ]
    var_ids = ["var1", "var2"]
    assert all_var_ids_in_records(records, var_ids) is True


def test_records_with_none():
    """
    Test when records contain None values.
    """
    records = [
        None,
        {
            "data": {
                "var": {
                    "var1": "value3",
                    "var2": "value4",
                }
            }
        },
    ]
    var_ids = ["var1", "var2"]
    with pytest.raises(ValueError):
        all_var_ids_in_records(records, var_ids)
